/*
 * Copyright (c) 2019-2020 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution  and  use  in  source  and binary  forms,  with  or  without
 * modification, are permitted provided that the following conditions are met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice and this list of conditions.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice and  this list of  conditions in the  documentation and/or
 *      other materials provided with the distribution.
 *
 * THE SOFTWARE  IS PROVIDED "AS IS"  AND THE AUTHOR  DISCLAIMS ALL WARRANTIES
 * WITH  REGARD   TO  THIS  SOFTWARE  INCLUDING  ALL   IMPLIED  WARRANTIES  OF
 * MERCHANTABILITY AND  FITNESS.  IN NO EVENT  SHALL THE AUTHOR  BE LIABLE FOR
 * ANY  SPECIAL, DIRECT,  INDIRECT, OR  CONSEQUENTIAL DAMAGES  OR  ANY DAMAGES
 * WHATSOEVER  RESULTING FROM  LOSS OF  USE, DATA  OR PROFITS,  WHETHER  IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR  OTHER TORTIOUS ACTION, ARISING OUT OF OR
 * IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 *                                           Anthony Mallet on Fri Jul  5 2019
 */
#include "autoconf.h"

#include "tk3-sys.h"


static void	tk3rc_icuwidth(ICUDriver *icup);
static void	tk3rc_icuovf(ICUDriver *icup);

static int16_t tk3rc_channels[8];
static unsigned int tk3rc_channel;


/* --- tk3rc_init ---------------------------------------------------------- */

static const ICUConfig tk3rc_icuconfig = {
  .mode =		ICU_INPUT_ACTIVE_HIGH,
  .channel =		ICU_CHANNEL_2,
  .frequency =		1000000,		/* 1MHz ICU clock frequency */
  .width_cb =		tk3rc_icuwidth,
  .overflow_cb =	tk3rc_icuovf
};

int
tk3rc_init()
{
  palSetPadMode(GPIOA, 1, PAL_MODE_ALTERNATE(2));
  icuStart(&ICUD5, &tk3rc_icuconfig);
  icuStartCapture(&ICUD5);
  icuEnableNotifications(&ICUD5);

  return 0;
}


/* --- tk3rc_channel_data -------------------------------------------------- */

const int16_t *
tk3rc_channel_data()
{
  return tk3rc_channels;
}


/* --- tk3rc_icuwidth ------------------------------------------------------ */

static void
tk3rc_icuwidth(ICUDriver *icup)
{
  icucnt_t w = icuGetWidthX(icup);
  int16_t c;

  if (tk3rc_channel >= sizeof(tk3rc_channels)/sizeof(tk3rc_channels[0]))
    tk3rc_channel = 0;
  if (w > 5000) {
    tk3rc_channel = 0;
    return;
  }

  if (w > 1500) w = 1500; else if (w < 700) w = 700;
  c = (w - 1100) * 81;

  with_syslock() {
    tk3rc_channels[tk3rc_channel++] = c;
  }
}


/* --- tk3rc_icuovf -------------------------------------------------------- */

static void
tk3rc_icuovf(ICUDriver *icup)
{
  tk3rc_channel = 0;

  with_syslock() {
    unsigned int i;

    for(i = 0; i < sizeof(tk3rc_channels)/sizeof(tk3rc_channels[0]); i++)
      tk3rc_channels[i] = 0;
  }
}
